/*
 * Sketchy -- An interpreter for purely applicative Scheme
 * Copyright (C) 2005,2006 Nils M Holm <nmh@t3x.org>
 * This is the interpreter shell.
 * See the file LICENSE for conditions of use.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include "sketchy.h"

#define STRLEN	256

char	Image[STRLEN];
int	Nodes, Vcells;
int	PrintVoid;

void	catchIntr(int sig);
int	getOptNum(int argc, char **argv, int *pi, int *pj, int *pk);
void	getOpts(int argc, char **argv);
void	help(void);
void	init(void);
void	repl(void);
void	usage(void);

/* Print usage and fail. */
void usage(void) {
	fprintf(stderr,
		"Usage: sketchy [-bi] [-n nodes] [-m vcells] [image]\n");
}

int getOptNum(int argc, char **argv, int *pi, int *pj, int *pk) {
	int	n, c;

	if (++(*pi) >= argc) {
		usage();
		exit(1);
	}
	n = atoi(argv[*pi]);
	c = argv[*pi][strlen(argv[*pi])-1];
	switch (c) {
	case 'K':	n = n * 1024; break;
	case 'M':	n = n * 1024 * 1024; break;
	}
	*pj = *pk = 0;
	return n;
}

void help(void) {
	fputc('\n', stderr);
	usage();
	fprintf(stderr,
		"\n"
		"-b    batch mode (quiet, exit on first error)\n"
		"-B    batch mode, but print void results\n"
		"-i    init mode (do not load any image)\n"
		"-m #  number of vector cells to allocate (default: %dK)\n"
		"-n #  number of nodes to allocate (default: %dK)\n"
		"\n"
		"default image: %s\n\n",
		SK_DFL_NODES/1024, SK_DFL_VCELLS/1024, SK_DFL_IMAGE);
}

/* Evaluate command line options. */
void getOpts(int argc, char **argv) {
	char	*a;
	int	i, j, k;

	strncpy(Image, SK_DFL_IMAGE, strlen(SK_DFL_IMAGE));
	Image[STRLEN-1] = 0;
	Nodes = SK_DFL_NODES;
	Vcells = SK_DFL_VCELLS;
	PrintVoid = 1;
	i = 1;
	while (i < argc) {
		a = argv[i];
		if (a[0] != '-') break;
		k = strlen(a);
		for (j=1; j<k; j++) {
			switch (a[j]) {
			case 'b':
				SK_batch = 1;
				PrintVoid = 0;
				break;
			case 'B':
				SK_batch = 1;
				PrintVoid = 1;
				break;
			case 'n':
				Nodes = getOptNum(argc, argv, &i, &j, &k);
				break;
			case 'm':
				Vcells = getOptNum(argc, argv, &i, &j, &k);
				break;
			case 'i':
				Image[0] = 0;
				break;
			case '?':
			case 'h':
				help();
				exit(1);
				break;
			default:
				usage();
				exit(1);
			}
		}
		i = i+1;
	}
	if (i < argc) {
		strncpy(Image, a, strlen(a));
		Image[STRLEN-1] = 0;
	}
	if (Nodes < SK_MIN_SIZE || Vcells < SK_MIN_SIZE) {
		fprintf(stderr, "sketchy: minimum pool size is %d\n",
			SK_MIN_SIZE);
		exit(1);
	}
}

void catchIntr(int sig) {
	sig = 0; /*LINT*/
	sk_stop();
	signal(SIGINT, catchIntr);
}

void repl(void) {
	int	n;

	while(1) {
		SK_errFlag = 0;
		n = sk_read();
		if (n == sk_eof()) return;
		if (SK_errFlag) {
			sk_printError();
			if (SK_batch) exit(1);
			continue;
		}
		n = sk_eval(n);
		if (SK_errFlag) {
			sk_printError();
			if (SK_batch) exit(1);
		}
		else {
			if (PrintVoid || n != sk_void()) {
				if (!SK_batch) sk_pr("=> ");
				sk_print(n); sk_nl();
			}
			if (SK_statFlag) sk_printStats();
		}
	}
}

void init(void) {
	if (sk_init(Nodes, Vcells)) {
		fprintf(stderr, "sketchy init failed (memory problem)\n");
		exit(1);
	}
}

/*
 * Here the fun begins.
 */
int main(int argc, char **argv) {
	getOpts(argc, argv);
	init();
	getOpts(argc, argv);
	if (!SK_batch) {
		sk_pr("Sketchy ");
		sk_pr(SK_RELEASE);
		sk_pr(" (C) 2006 Nils M Holm"); sk_nl();
	}
	if (Image[0]) {
		if (sk_loadImage(Image)) {
			sk_printError();
			if (SK_batch) exit(1);
			sk_fini();
			init();
			getOpts(argc, argv);
		}
	}
	else if (!SK_batch) {
		sk_pr("Warning: no image loaded"); sk_nl();
	}
	signal(SIGINT, catchIntr);
	repl();
	sk_bye();
	return 0;
}
